#!/usr/bin/perl
#AIM Sniff Copyright (C) 2002 Shawn Grimes
#This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as 
#published by the Free Software Foundation; version 2 of the License.
# This program is distributed in the hope that it will be useful, but 
#WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
#or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
#more details.
#You should have received a copy of the GNU General Public License along with 
#this program; if not, write to the 
#Free Software Foundation, Inc.
#59 Temple Place, Suite 330
#Boston, MA 02111-1307 USA

#You may also contact me directly with any questions at: 
#grimessh@users.sourceforge.net

use Net::Pcap;
use NetPacket::Ethernet qw(:strip);
use NetPacket::IP qw(:strip);
use NetPacket::TCP;
use DBI;
use Unicode::String qw(utf8 latin1 utf16);

#use Net::RawIP;
#use Net::RawIP::libpcap;


######################################################################
#########   User Defined Variables                           #########
######################################################################
$debug=0;
$driver="mysql";
$database="aim";
$host="mysql.server.com";
$user="username";
$password="password";
######################################################################
#########   End User Defined Variables--Modify nothing below #########
######################################################################

$ver='0.4';

####
#Get command line arguments
####
&getOptions(@ARGV) || die "Could not get options\n";



if(defined($getHandles)){
    &getFromHandle || die "Could not get SMB Names\n"; ##Used to find out who messages are coming from in the database.  
}else{
    if(defined($dumpfile)){ #work in offline mode?
	print "Working in Offline Mode\n******************\nReading File: $dumpfile\n\n";
	$pcap_t = Net::Pcap::open_offline($dumpfile, \$err);
	print "Error opening file: $err\n" if(!defined($pcap_t));

    }else{  #actively sniff
	print "Beginning Sniff:\n*************\n\n" if(!$quiet);
	####Command Line Options
	$dev='eth0' if(!defined($dev));
	$filter_str='tcp and port 5190' if(!defined($filter_str));
	$promisc=1 if(!defined($promisc));
	$to_ms=1000 if(!defined($to_ms));

	###These shouldn't be changed (yet)
	$optimize=0;
	$netmask=0;
	$snaplen=50000;
	$pcap_t=Net::Pcap::open_live($dev, $snaplen, $promisc, $to_ms, \$err) || die "Error opening pcap: $err\n"; #start sniffing
	$result=Net::Pcap::compile($pcap_t, \$filter_t, $filter_str,$optimize,$netmask); #compile filter_str
	Net::Pcap::setfilter($pcap_t, $filter_t); #apply filter
    }
    Net::Pcap::loop($pcap_t, $count, \&process_pkt,"xyz"); #start to process the packets that are received
    Net::Pcap::stats($pcap_t, \%stats);
    print "Packets Received: $stats->{ps_recv}\nPackets Dropped: $stats->{ps_drop}\n"; #print stats but only works when $count is defined, I'd like to find a way to print it out when CTRL-C is used to quit the program on a continuous loop
}



#Process the packet
sub process_pkt {
    my($user, $hdr, $pkt) = @_;
    if (!defined($hdr) or !defined($pkt)) {
	print("Bad args passed to callback\n");
	print("Bad user data\n"), if ($user ne "xyz");
	print("Bad pkthdr\n"), if (!defined($hdr));
	print("Bad pkt data\n"), if (!defined($pkt));
	print("not ok\n");
	exit;
    }
#get datetimestamp of packet
    ($sec, $min, $hour, $mday, $mon, $year)=localtime($hdr->{tv_sec});
    $year+=1900;
    $mon+=1;	
    $datestamp="$year-$mon-$mday $hour:$min:$sec";

#Strip Ethernet portion of packet off
    $ip_obj=NetPacket::IP->decode(eth_strip($pkt));
    $srcip=$ip_obj->{src_ip};
    $dstip=$ip_obj->{dest_ip};
    $tcp_obj=NetPacket::TCP->decode(ip_strip(eth_strip($pkt)));
    $flags=$tcp_obj->{flags};  #make sure packet contains ACK PUSH

    $dataset=$tcp_obj->{data};

    $family=&familyFind($dataset); #get Family ID and SubID
    if(($family == '00040006' or $family=='00040007') and $flags == '24'){ #AIM Message
	$handle=&handleFind($dataset); #Get AIM Handle
	$msg=&msgFind($dataset); #Get the message
	$msg=&msgClean($msg);  #Strip HTML and the other junk off the message
	if($family == '00040007' and $handle and $msg){
	    &mysqlAdd("logs",$datestamp, $handle, $faimly, $msg, $dstip) if(!$nodb);
	    &printOUT($datestamp, $handle, "INCOMING", $msg, $dstip) if(!$quiet);
	}else{
	    &mysqlAdd("logs",$datestamp, $handle, $family, $msg, $srcip) if(!$nodb and $handle and $msg);
	    &printOUT($datestamp, $handle, "OUTGOING", $msg, $srcip) if(!$quiet and $handle and $msg);
	}

    }elsif($family == '00170006' and defined($SMB)){ #AIM Login
	$handle=&getSignon($dataset); #Get AIM Handle
	($SMBMachine,$SMBUser)=&SMBInfo($srcip); #Get SMB username and computername if available
	print "\nAIM Login:\n$SMBUser on $SMBMachine logged on as $handle\n\n" if(!($quiet) and ($SMBMachine) and ($SMBUser)); #print to STDOUT
	&mysqlAdd("handles",$datestamp, $handle, $SMBUser, $SMBMachine, $srcip) if(!$nodb and $SMBMachine and $SMBUser); #add to Database
    }
}

sub familyFind{
#Find family id
	my($dataset)=@_;
        $family="";
        if(length($dataset) > 0){
		$family=utf8($dataset);
		$family=$family->hex;
		$family=~s/U\+00//g;
		if ($family=~/(00 04 00 0[67])/ or $family=~/(00 17 00 06)/){
		    $family=$1;
		    $family=~s/\s//g;
		}else{
		    $family=""
		}
	 }
	return $family;
}

sub getSignon{
#Find AIM Handle
    my($dataset)=@_;
    $handle="";
    $dataset=utf8($dataset);
    $dataset=$dataset->hex;
    $dataset=~s/U\+00//g;
    $dataset=~s/.*00\s17\s00\s06\s([abcdef0123456789][abcdef0123456789]\s){10}//;
    $dataset=~s/\s00\s4b\s00\s00//;
    @digis=split(/\s/,$dataset);
    foreach $digi(@digis){
	$te=chr(hex($digi));
        $handle.=$te;
    }
    return $handle;
}

sub SMBInfo{
#Get SMB machine name and username
    my($ip)=@_;
    @SMBInfo=`nmblookup -A $ip`;
    if($SMBInfo[1]!~/ACTIVE/){
	print "Didn't find nmblookup entry: $SMBInfo[1]\n" if $debug;
	$SMBName="";
	$username="";
    }else{
	$SMBName=$SMBInfo[1];
	$SMBName=~/(.*)\<.*/;
	$SMBName=$1;
	$SMBName=~s/\s//;
	$SMBName=~s/\<.*//;
	$SMBName=utf8($SMBName);
	$SMBName=$SMBName->hex;
	$SMBName=~s/U\+00//g;
	$SMBName=~s/\s20//g;
	@digis=split(/\s/,$SMBName);
	$SMBName="";
	foreach $digi(@digis){
	    $te=chr(hex($digi));
	    $SMBName.=$te;
	}
			    
	##Get user name
	$username=$SMBInfo[$SMBInfo-2];
	$username=~/(.*)\s*\<03\>.*/;
	$username=$1;
	$username=~s/\s//;
	$username=utf8($username);
	$username=$username->hex;
	$username=~s/U\+00//g;
	$username=~s/\s20//g;
	@digis=split(/\s/,$username);
	$username="";
	foreach $digi(@digis){
	    $te=chr(hex($digi));
	    $username.=$te;
	}
    }
    return $SMBName, $username;
}
    

sub handleFind{
#Find AIM handle
    my($dataset)=@_;
    $handle="";
    $dataset=utf8($dataset);
    $dataset=$dataset->hex;
    $dataset=~s/U\+00//g;
    $dataset=~/.*00\s01\s0[a-f]\s([4-7][a-f0-9].*)\s00\s.*/;
    $dataset=$1;
    $dataset=~s/00\s.*//;
    @digis=split(/\s/,$dataset);
    foreach $digi(@digis){
	$te=chr(hex($digi));
	$handle.=$te;
    }
    return $handle;
}

sub msgFind{
#Find AIM msg
	my($dataset)=@_;
	$msg="";
	$dataset=utf8($dataset);
	$dataset=$dataset->hex;
	$dataset=~s/U\+00//g;
        $hexcomb=$dataset;
        $hexcomb=~s/(.*00\s00\s00\s00\s)//;
        $hexcomb=~s/(00\s08.*)//;
        $msg="";
        @digis=split(/\s/,$hexcomb);
        foreach $digi(@digis){
        	$te=chr(hex($digi));
                $msg.=$te;
        }
	return $msg;
}

sub msgClean{
#Remove HTML tags from message
    my($msg)=@_;
    $msg=~s/<([^>]|\n)*>//g;
    $msg=~s/\<\/.*//;
    return $msg;
}

sub mysqlAdd{
    my($datestamp, $handle, $family, $msg)=@_;
    my(@values)=@_;
    my($table)=shift(@values);

    $dsn="DBI:$driver:database=$database:host=$host:user=$user:password=$password";
    $dbh=DBI->connect($dsn) || return 0;
    $dbh->{RaiseError}=0;
    
    if($table=~/logs/){
	$sql="INSERT INTO logs SET ";
	foreach $value(@values){
	    $value=$dbh->quote($value);
	    $value=~s/\0//;
	}
	my($datestamp, $handle, $direction, $msg, $ip)=@values;
	$sql.="ts=$datestamp, handle=$handle, direction=$direction, message=$msg, ip=$ip";
    }elsif($table=~/handles/){
	$sql="INSERT INTO handles SET ";
	foreach $value(@values){
	    $value=$dbh->quote($value);
	}
	my($datestamp, $handle, $user, $machine, $ip)=@values;
	$sql.="ts=$datestamp, handle=$handle, username=$user, machine=$machine, ip=$ip";
    }

    $sth=$dbh->prepare($sql) || print "problem preparing: ",$dbh->errstr,"\n";
    if($handle){
	$sth->execute || print "Error executing $DBI::errstr\n";
	print "SQL: $sql\n" if $debug;
    }
}

sub printOUT{
    my($datestamp, $handle, $family, $msg, $ip)=@_;
    print "DATE STAMP: $datestamp\n";
    print "IP: $ip\n";
    print "DIRECTION: $family\n";
    print "HANDLE: $handle\n";
    print "MESSAGE: $msg\n\n";
}

sub getFromHandle{
    $dsn="DBI:$driver:database=$database:host=$host:user=$user:password=$password";
    $dbh=DBI->connect($dsn) || return 0;
    $dbh->{RaiseError}=0;
   
    $sql="SELECT id, ip, ts FROM logs WHERE direction='00040006' and fromHandle is NULL or fromHandle=''";
    $sth=$dbh->prepare($sql) || die "problem preparing: ",$dbh->errstr,"\n";
    $sth->execute || die "Error executing $DBI::errstr\n";
    $sth->bind_col(1, \$id);
    $sth->bind_col(2, \$ip);
    $sth->bind_col(3, \$ts);
    while(@values=$sth->fetchrow_array()){
	$handle="";
	$sql2="SELECT handle FROM handles WHERE ip='$ip' AND ts<'$ts' order by ts desc limit 1";
	$sth2=$dbh->prepare($sql2) || die "Problem preparing: ",$dbh->errstr,"\n";
	$sth2->execute || die "Error executing $DBI::errstr\n";
	$sth2->bind_col(1, \$handle);
	@row=$sth2->fetchrow_array();
	if($handle){
	    $sql2="UPDATE logs SET fromHandle='$handle' WHERE id='$id'";
	    $sth2=$dbh->prepare($sql2) || die "Problem preparing: ",$dbh->errstr,"\n";
	    $sth2->execute || die "Error executing $DBI::errstr\n";
	}
    }
}


sub getOptions{
    my(@ARGV)=@_;
    if($ARGV[0]=~/-C=(.*)/){
	$configfile=$1;
	open(CONFIG,"$configfile") || print "Error opening config file, using defaults instead.\n";
	while(<CONFIG>){
	    $arg=$_;
	    if($arg!~/^#/){
	       $dumpfile=$1 if($arg=~/dumpfile=(.*)/);
	       $count=$1 if($arg=~/packetCount=(.*)/);
	       $dev=$1 if($arg=~/dev=(.*)/);
	       $filter_str=$1 if($arg=~/filter='(.*)'/);
	       $promisc=1 if ($arg=~/promisc=1/);
	       $to_ms=$1 if ($arg=~/timeout=(.*)/);
	       $nodb=1 if ($arg=~/nodb=1/);
	       $quiet=1 if ($arg=~/quiet=1/);
	       $SMB=1 if ($arg=~/SMB=1/);
	       $host=$1 if ($arg=~/host=(.*)/);
	       $user=$1 if ($arg=~/user=(.*)/);
	       $password=$1 if ($arg=~/password=(.*)/);
	   }
	}
	print "SECOND ARG IS: $ARGV[1]\n";
	$getHandles=1 if($ARGV[1]=~/--getHandles/);
	print "Running with: \tdumpfile=$dumpfile\n\tpacketCount=$count\n\tdev=$dev\n\tfilter=$filter_str\n\tpromisc=$promisc\n\ttimeout=$to_ms\n\tSMB=$SMB\n\tnodb=$nodb\n\tquiet=$quiet\n\thost=$host\n\tuser=$user\n\tpassword=****\n\n";

    }else{
	foreach $arg (@ARGV){
	    $getHandles=1 if($arg=~/--getHandles/);
	    $dumpfile=$1 if($arg=~/-r=(.*)/);
	    $count=$1 if($arg=~/-c=(.*)/);
	    $dev=$1 if($arg=~/-d=(.*)/);
	    $filter_str=$1 if($arg=~/-f='(.*)'/);
	    $promisc=1 if ($arg=~/-p/);
	    $to_ms=$1 if ($arg=~/-to=(.*)/);
	    $nodb=1 if ($arg=~/--nodb/);
	    $quiet=1 if ($arg=~/--quiet/);
	    $SMB=1 if ($arg=~/--SMB/);
	    if($arg=~/--h/ or $arg=~/-h/){
		print "**************\nAimSniff $ver\n**************\n";
		print "AIM Sniff Copyright (C) 2002 Shawn Grimes\n";
		print "This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; version 2.\n";
		print "This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.\n";
		print "You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA\n";
		print "You may also contact me directly with any questions at: grimessh\@users.sourceforge.net\n";
		print "\n###########\nTo use:\n";
		print "\t-C=filename <-Get AIM Sniff options from a config file\n";
		print "\t-r=filename <-Read a PCAP file instead of doing a live capture\n";
		print "\t-c=integer <-The number of packets to read before quitting\n";
		print "\t-d=dev <-The device to capture packets from\n";
		print "\t-f='filter string' <-String to filter on enclosed in single quotes\n\t\t(DEFAULT: 'tcp and port 5190') -- Should only have to be specified if you think AIM is running on a different port\n";
		print "\t-p <-Place the device into promiscuous mode\n";
		print "\t-to=integer <-Read timeout in ms\n";
		print "\t--SMB <-Turn SMB lookups 'on' to get NT domain usernames with AIM logins, Off by default\n";
		print "\t--nodb <-Do not dump to a DB, only dump to STDOUT\n";
		print "\t--quiet <-Do not print anything but errors to STDOUT\n";
		print "\t--getHandles <-Do not do anything with PCAP but populate the fromHandle field in the logs table\n";
		print "\nDefaults:\n";
		print "\t-d=eth0\n\t-f='tcp and port 5190'\n\t-p\n\t-to=1000\n";
		exit;
	    }
	}
    }
    return 1;
}
